#!/usr/bin/env python


import unittest

from stone.backend import (
    remove_aliases_from_api,
    CodeBackend
)
from stone.ir import (
    Alias,
    Api,
    ApiNamespace,
    ApiRoute,
    List,
    Nullable,
    Boolean,
    String,
    Struct,
    StructField,
    Union,
    UnionField,
    resolve_aliases,
    strip_alias
)

_MYPY = False
if _MYPY:
    import typing  # noqa: F401 # pylint: disable=import-error,unused-import,useless-suppression

import argparse
class _Tester(CodeBackend):
    """A no-op backend used to test helper methods."""
    def generate(self, api):
        pass

class _TesterCmdline(CodeBackend):
    cmdline_parser = argparse.ArgumentParser()
    cmdline_parser.add_argument('-v', '--verbose', action='store_true')
    def generate(self, api):
        pass

class TestBackend(unittest.TestCase):
    """
    Tests the interface exposed to backends.
    """

    def test_api_namespace(self):
        ns = ApiNamespace('files')
        a1 = Struct('A1', None, ns)
        a1.set_attributes(None, [StructField('f1', Boolean(), None, None)])
        a2 = Struct('A2', None, ns)
        a2.set_attributes(None, [StructField('f2', Boolean(), None, None)])
        l1 = List(a1)
        s = String()
        route = ApiRoute('test/route', 1, None)
        route.set_attributes(None, None, l1, a2, s, None)
        ns.add_route(route)

        # Test that only user-defined types are returned.
        route_io = ns.get_route_io_data_types()
        self.assertIn(a1, route_io)
        self.assertIn(a2, route_io)
        self.assertNotIn(l1, route_io)
        self.assertNotIn(s, route_io)

    def test_code_backend_helpers(self):
        t = _Tester(None, [])
        self.assertEqual(t.filter_out_none_valued_keys({}), {})
        self.assertEqual(t.filter_out_none_valued_keys({'a': None}), {})
        self.assertEqual(t.filter_out_none_valued_keys({'a': None, 'b': 3}), {'b': 3})

    def test_code_backend_basic_emitters(self):
        t = _Tester(None, [])

        # Check basic emit
        t.emit('hello')
        self.assertEqual(t.output_buffer_to_string(), 'hello\n')
        t.clear_output_buffer()

        # Check that newlines are disallowed in emit
        self.assertRaises(AssertionError, lambda: t.emit('hello\n'))

        # Check indent context manager
        t.emit('hello')
        with t.indent():
            t.emit('world')
            with t.indent():
                t.emit('!')
        expected = """\
hello
    world
        !
"""
        self.assertEqual(t.output_buffer_to_string(), expected)
        t.clear_output_buffer()

        # --------------------------------------------------------
        # Check text wrapping emitter

        with t.indent():
            t.emit_wrapped_text('Colorless green ideas sleep furiously',
                                prefix='$', initial_prefix='>',
                                subsequent_prefix='|', width=13)
        expected = """\
    $>Colorless
    $|green
    $|ideas
    $|sleep
    $|furiously
"""
        self.assertEqual(t.output_buffer_to_string(), expected)
        t.clear_output_buffer()

    def test_code_backend_list_gen(self):
        t = _Tester(None, [])

        t.generate_multiline_list(['a=1', 'b=2'])
        expected = """\
(a=1,
 b=2)
"""
        self.assertEqual(t.output_buffer_to_string(), expected)
        t.clear_output_buffer()

        t.generate_multiline_list(['a=1', 'b=2'], 'def __init__', after=':')
        expected = """\
def __init__(a=1,
             b=2):
"""
        self.assertEqual(t.output_buffer_to_string(), expected)
        t.clear_output_buffer()

        t.generate_multiline_list(['a=1', 'b=2'], before='function_to_call', compact=False)
        expected = """\
function_to_call(
    a=1,
    b=2,
)
"""
        self.assertEqual(t.output_buffer_to_string(), expected)
        t.clear_output_buffer()

        t.generate_multiline_list(['a=1', 'b=2'], 'function_to_call',
                                  compact=False, skip_last_sep=True)
        expected = """\
function_to_call(
    a=1,
    b=2
)
"""
        self.assertEqual(t.output_buffer_to_string(), expected)
        t.clear_output_buffer()

        t.generate_multiline_list(['a=1', 'b=2'], 'def func', ':',
                                  compact=False, skip_last_sep=True)
        expected = """\
def func(
    a=1,
    b=2
):
"""
        self.assertEqual(t.output_buffer_to_string(), expected)
        t.clear_output_buffer()

        t.generate_multiline_list(['a=1'], 'function_to_call', compact=False)
        expected = 'function_to_call(a=1)\n'
        self.assertEqual(t.output_buffer_to_string(), expected)
        t.clear_output_buffer()

        t.generate_multiline_list(['a=1'], 'function_to_call', compact=True)
        expected = 'function_to_call(a=1)\n'
        self.assertEqual(t.output_buffer_to_string(), expected)
        t.clear_output_buffer()

        t.generate_multiline_list([], 'function_to_call', compact=False)
        expected = 'function_to_call()\n'
        self.assertEqual(t.output_buffer_to_string(), expected)
        t.clear_output_buffer()

        t.generate_multiline_list([], 'function_to_call', compact=True)
        expected = 'function_to_call()\n'
        self.assertEqual(t.output_buffer_to_string(), expected)
        t.clear_output_buffer()

        # Test delimiter
        t.generate_multiline_list(['String'], 'List', delim=('<', '>'), compact=True)
        expected = 'List<String>\n'
        self.assertEqual(t.output_buffer_to_string(), expected)
        t.clear_output_buffer()

    def test_code_backend_block_gen(self):
        t = _Tester(None, [])

        with t.block('int sq(int x)', ';'):
            t.emit('return x*x;')
        expected = """\
int sq(int x) {
    return x*x;
};
"""
        self.assertEqual(t.output_buffer_to_string(), expected)
        t.clear_output_buffer()

        with t.block('int sq(int x)', allman=True):
            t.emit('return x*x;')
        expected = """\
int sq(int x)
{
    return x*x;
}
"""
        self.assertEqual(t.output_buffer_to_string(), expected)
        t.clear_output_buffer()

        with t.block('int sq(int x)', delim=('<', '>'), dent=8):
            t.emit('return x*x;')
        expected = """\
int sq(int x) <
        return x*x;
>
"""
        self.assertEqual(t.output_buffer_to_string(), expected)
        t.clear_output_buffer()

    def test_backend_cmdline(self):
        t = _TesterCmdline(None, ['-v'])
        self.assertTrue(t.args.verbose)

    def test_resolve_aliases(self):
        first_alias = Alias(None, None, None)
        first_alias.data_type = String()
        resolved_type = resolve_aliases(first_alias.data_type)

        # Test that single-level alias chain resolves
        self.assertIsInstance(resolved_type, String)
        self.assertIsInstance(first_alias.data_type, String)

        first_alias = Alias(None, None, None)
        second_alias = Alias(None, None, None)
        first_alias.data_type = second_alias
        second_alias.data_type = String()

        # Test that a two-level alias chain resolves
        resolved_type = resolve_aliases(first_alias.data_type)
        first_alias.data_type = resolved_type

        self.assertIsInstance(resolved_type, String)
        self.assertIsInstance(first_alias.data_type, String)
        self.assertIsInstance(second_alias.data_type, String)

    def test_strip_alias(self):
        first_alias = Alias(None, None, None)
        second_alias = Alias(None, None, None)
        third_alias = Alias(None, None, None)
        first_alias.data_type = second_alias
        second_alias.data_type = third_alias
        third_alias.data_type = String()

        test_struct = Struct('TestStruct', None, None)
        test_struct.set_attributes(None, [
            StructField('field1', List(Nullable(first_alias)), None, None)
        ])

        curr_type = first_alias
        while hasattr(curr_type, 'data_type'):
            curr_type.data_type = resolve_aliases(curr_type.data_type)
            curr_type = curr_type.data_type

        self.assertIsInstance(first_alias.data_type, String)

        self.assertEqual(len(test_struct.fields), 1)
        field = test_struct.fields[0]

        strip_alias(field.data_type)

        list_type = field.data_type
        self.assertIsInstance(list_type, List)
        nullable_type = list_type.data_type
        self.assertIsInstance(nullable_type, Nullable)
        string_type = nullable_type.data_type
        self.assertIsInstance(string_type, String)

    def test_preserve_aliases_from_api(self):
        api = Api(version="0.1b1")
        # Ensure imports come after 'preserve_alias' lexiographicaly
        # to catch namespace ordering bugs
        api.ensure_namespace('preserve_alias')
        api.ensure_namespace('zzzz')

        ns = api.namespaces['preserve_alias']
        imported = api.namespaces['zzzz']

        namespace_id = Alias('NamespaceId', ns, None)
        namespace_id.data_type = String()
        shared_folder_id = Alias('SharedFolderId', ns, None)
        shared_folder_id.set_attributes(None, namespace_id)
        path_root_id = Alias('PathRootId', ns, None)
        path_root_id.set_attributes(None, shared_folder_id)
        foo_alias = Alias('FooAlias', None, None)
        foo_alias.set_attributes(None, String())
        bar_alias = Alias('BarAlias', None, None)
        bar_alias.set_attributes(None, foo_alias)

        ns.add_alias(namespace_id)
        ns.add_alias(shared_folder_id)
        ns.add_alias(path_root_id)
        imported.add_alias(foo_alias)
        imported.add_alias(bar_alias)

        test_struct = Struct('TestStruct', ns, None)
        test_struct.set_attributes(None, [StructField('field1', path_root_id, None, None)])
        test_union = Union('TestUnion', ns, None, None)
        test_union.set_attributes(None, [UnionField('test', path_root_id, None, None)])
        dependent_struct = Struct('DependentStruct', ns, None)
        dependent_struct.set_attributes(None, [
            StructField('field_alias', imported.alias_by_name['BarAlias'], None, None)
        ])

        ns.add_data_type(test_struct)
        ns.add_data_type(test_union)
        ns.add_data_type(dependent_struct)

        struct_alias = Alias('StructAlias', ns, None)
        struct_alias.set_attributes(None, test_struct)

        ns.add_alias(struct_alias)

        # Ensure namespace exists
        self.assertEqual(len(api.namespaces), 2)
        self.assertTrue('preserve_alias' in api.namespaces)
        self.assertTrue('zzzz' in api.namespaces)

        ns = api.namespaces['preserve_alias']
        imported = api.namespaces['zzzz']

        # Ensure aliases exist
        self.assertEqual(len(ns.aliases), 4)
        self.assertEqual(len(imported.aliases), 2)

        aliases = {
            alias._name: alias for alias in ns.aliases
        }
        imported_aliases = {
            alias.name: alias for alias in imported.aliases
        }
        data_types = {
            data_type._name: data_type for data_type in ns.data_types
        }

        # Ensure aliases are in the namespace
        self.assertTrue('NamespaceId' in aliases)
        self.assertTrue('SharedFolderId' in aliases)
        self.assertTrue('PathRootId' in aliases)
        self.assertTrue('StructAlias' in aliases)
        self.assertTrue('FooAlias' in imported_aliases)
        self.assertTrue('BarAlias' in imported_aliases)

        # Ensure aliases resolve to proper types
        self.assertIsInstance(aliases['NamespaceId'].data_type, String)
        self.assertIsInstance(aliases['SharedFolderId'].data_type, Alias)
        self.assertIsInstance(aliases['PathRootId'].data_type, Alias)
        self.assertIsInstance(aliases['StructAlias'].data_type, Struct)
        self.assertIsInstance(imported_aliases['FooAlias'].data_type, String)
        self.assertIsInstance(imported_aliases['BarAlias'].data_type, Alias)

        # Ensure struct and union field aliases resolve to proper types
        self.assertIsInstance(data_types['TestStruct'], Struct)

        test_struct = data_types.get('TestStruct')
        dependent_struct = data_types.get('DependentStruct')

        self.assertTrue(len(test_struct.fields), 1)
        self.assertTrue(len(dependent_struct.fields), 1)

        field = test_struct.fields[0]
        self.assertEqual(field.name, 'field1')
        self.assertIsInstance(field.data_type, Alias)

        field = dependent_struct.fields[0]
        self.assertEqual(field.name, 'field_alias')
        self.assertIsInstance(field.data_type, Alias)

        test_union = data_types['TestUnion']

        self.assertTrue(len(test_union.fields), 1)
        field = test_union.fields[0]

        self.assertEqual(field.name, 'test')
        self.assertIsInstance(field.data_type, Alias)

    def test_no_preserve_aliases_from_api(self):
        api = Api(version="0.1b1")
        # Ensure imports come after 'preserve_alias' lexiographicaly
        # to catch namespace ordering bugs
        api.ensure_namespace('preserve_alias')
        api.ensure_namespace('zzzz')

        ns = api.namespaces['preserve_alias']
        imported = api.namespaces['zzzz']

        # Setup aliases
        namespace_id = Alias('NamespaceId', ns, None)
        namespace_id.data_type = String()
        shared_folder_id = Alias('SharedFolderId', ns, None)
        shared_folder_id.set_attributes(None, namespace_id)
        path_root_id = Alias('PathRootId', ns, None)
        path_root_id.set_attributes(None, shared_folder_id)
        nullable_alias = Alias('NullableAlias', ns, None)
        nullable_alias.set_attributes(None, Nullable(path_root_id))
        foo_alias = Alias('FooAlias', None, None)
        foo_alias.set_attributes(None, String())
        bar_alias = Alias('BarAlias', None, None)
        bar_alias.set_attributes(None, foo_alias)

        ns.add_alias(namespace_id)
        ns.add_alias(shared_folder_id)
        ns.add_alias(path_root_id)
        ns.add_alias(nullable_alias)
        imported.add_alias(foo_alias)
        imported.add_alias(bar_alias)

        # Setup composite types
        test_struct = Struct('TestStruct', ns, None)
        test_struct.set_attributes(None, [
            StructField('field_alias', path_root_id, None, None),
            StructField('field_nullable_alias', nullable_alias, None, None),
            StructField('field_list_of_alias', List(path_root_id), None, None)
        ])
        test_union = Union('TestUnion', ns, None, None)
        test_union.set_attributes(None, [UnionField('test', path_root_id, None, None)])
        dependent_struct = Struct('DependentStruct', ns, None)
        dependent_struct.set_attributes(None, [
            StructField('field_alias', imported.alias_by_name['BarAlias'], None, None)
        ])

        ns.add_data_type(test_struct)
        ns.add_data_type(test_union)
        ns.add_data_type(dependent_struct)

        # Setup aliases on composite types
        struct_alias = Alias('StructAlias', ns, None)
        struct_alias.set_attributes(None, test_struct)

        ns.add_alias(struct_alias)

        api = remove_aliases_from_api(api)

        # Ensure namespace exists
        self.assertEqual(len(api.namespaces), 2)
        self.assertTrue('preserve_alias' in api.namespaces)
        self.assertTrue('zzzz' in api.namespaces)

        ns = api.namespaces['preserve_alias']
        imported = api.namespaces['zzzz']

        # Ensure aliases are gone
        self.assertEqual(len(ns.aliases), 0)
        self.assertEqual(len(imported.aliases), 0)

        data_types = {
            data_type._name: data_type for data_type in ns.data_types
        }

        # Ensure struct and union field aliases resolve to proper types
        test_struct = data_types.get('TestStruct')
        self.assertIsInstance(test_struct, Struct)

        self.assertEqual(len(test_struct.fields), 3)

        for field in test_struct.fields:
            if field.name == 'field_list_of_alias':
                self.assertIsInstance(field.data_type, List)
                list_type = field.data_type.data_type
                self.assertIsInstance(list_type, String)
            elif field.name == 'field_nullable_alias':
                field_type = field.data_type
                self.assertIsInstance(field_type, Nullable)
                self.assertIsInstance(field_type.data_type, String)
            else:
                self.assertIsInstance(field.data_type, String)

        test_union = data_types['TestUnion']

        self.assertTrue(len(test_union.fields), 1)
        field = test_union.fields[0]

        self.assertEqual(field.name, 'test')
        self.assertIsInstance(field.data_type, String)

        # Ensure struct using imported alias resolves properly
        dependent_struct = data_types.get('DependentStruct')
        self.assertIsInstance(dependent_struct, Struct)

        self.assertEqual(len(dependent_struct.fields), 1)

        field = dependent_struct.fields[0]
        self.assertIsInstance(field.data_type, String)

    def test_route_hashing(self):
        route = ApiRoute('test/route', 1, None)
        same_route = ApiRoute("test/route", 1, None)
        diff_version = ApiRoute("test/route", 2, None)
        diff_route = ApiRoute("test/route_2", 1, None)
        self.assertEqual(hash(route), hash(same_route))
        self.assertNotEqual(hash(route), hash(diff_version))
        self.assertNotEqual(hash(route), hash(diff_route))

if __name__ == '__main__':
    unittest.main()



from contextlib import contextmanager

from stone.backend import CodeBackend
from stone.backends.obj_c_helpers import (
    fmt_camel_upper,
    fmt_class,
    fmt_class_prefix,
    fmt_import, )
from stone.ir import (
    is_list_type,
    is_map_type,
    is_struct_type,
    is_union_type,
    is_nullable_type,
    is_user_defined_type,
    is_void_type,
    unwrap_nullable, )

stone_warning = """\
///
/// Copyright (c) 2016 Dropbox, Inc. All rights reserved.
///
/// Auto-generated by Stone, do not modify.
///

"""

# This will be at the top of the generated file.
base_file_comment = """\
{}\
""".format(stone_warning)

undocumented = '(no description).'

comment_prefix = '/// '


class ObjCBaseBackend(CodeBackend):
    """Wrapper class over Stone generator for Obj C logic."""
    # pylint: disable=abstract-method

    @contextmanager
    def block_m(self, class_name):
        with self.block(
                '@implementation {}'.format(class_name),
                delim=('', '@end'),
                dent=0):
            self.emit()
            yield

    @contextmanager
    def block_h_from_data_type(self, data_type, protocol=None):
        assert is_user_defined_type(data_type), \
            'Expected user-defined type, got %r' % type(data_type)

        if not protocol:
            extensions = []
            if data_type.parent_type and is_struct_type(data_type):
                extensions.append(fmt_class_prefix(data_type.parent_type))
            else:
                if is_union_type(data_type):
                    # Use a handwritten base class
                    extensions.append('NSObject')
                else:
                    extensions.append('NSObject')

            extend_suffix = ' : {}'.format(
                ', '.join(extensions)) if extensions else ''
        else:
            base = fmt_class_prefix(data_type.parent_type) if (
                data_type.parent_type and
                not is_union_type(data_type)) else 'NSObject'
            extend_suffix = ' : {} <{}>'.format(base, ', '.join(protocol))
        with self.block(
                '@interface {}{}'.format(
                    fmt_class_prefix(data_type), extend_suffix),
                delim=('', '@end'),
                dent=0):
            self.emit()
            yield

    @contextmanager
    def block_h(self,
                class_name,
                protocol=None,
                extensions=None,
                protected=None):
        if not extensions:
            extensions = ['NSObject']

        if not protocol:
            extend_suffix = ' : {}'.format(', '.join(extensions))
        else:
            extend_suffix = ' : {} <{}>'.format(', '.join(extensions),
                                                fmt_class(protocol))

        base_interface_str = '@interface {}{} {{' if protected else '@interface {}{}'
        with self.block(
                base_interface_str.format(class_name, extend_suffix),
                delim=('', '@end'),
                dent=0):
            if protected:
                with self.block('', delim=('', '')):
                    self.emit('@protected')
                    for field_name, field_type in protected:
                        self.emit('{} _{};'.format(field_type, field_name))
                self.emit('}')
            self.emit()
            yield

    @contextmanager
    def block_init(self):
        with self.block('if (self)'):
            yield
        self.emit('return self;')

    @contextmanager
    def block_func(self, func, args=None, return_type='void',
                   class_func=False):
        args = args if args is not None else []
        modifier = '-' if not class_func else '+'
        base_string = '{} ({}){}:{}' if args else '{} ({}){}'
        signature = base_string.format(modifier, return_type, func, args)
        with self.block(signature):
            yield

    def _get_imports_m(self, data_types, default_imports):
        """Emits all necessary implementation file imports for the given Stone data type."""
        if not isinstance(data_types, list):
            data_types = [data_types]

        import_classes = default_imports

        for data_type in data_types:
            import_classes.append(fmt_class_prefix(data_type))

            if data_type.parent_type:
                import_classes.append(fmt_class_prefix(data_type.parent_type))

            if is_struct_type(
                    data_type) and data_type.has_enumerated_subtypes():
                for _, subtype in data_type.get_all_subtypes_with_tags():
                    import_classes.append(fmt_class_prefix(subtype))

            for field in data_type.all_fields:
                data_type, _ = unwrap_nullable(field.data_type)

                # unpack list or map
                while is_list_type(data_type) or is_map_type(data_type):
                    data_type = (data_type.value_data_type if
                        is_map_type(data_type) else data_type.data_type)

                if is_user_defined_type(data_type):
                    import_classes.append(fmt_class_prefix(data_type))

        if import_classes:
            import_classes = list(set(import_classes))
            import_classes.sort()

        return import_classes

    def _get_imports_h(self, data_types):
        """Emits all necessary header file imports for the given Stone data type."""
        if not isinstance(data_types, list):
            data_types = [data_types]

        import_classes = []

        for data_type in data_types:

            if is_user_defined_type(data_type):
                import_classes.append(fmt_class_prefix(data_type))

            for field in data_type.all_fields:
                data_type, _ = unwrap_nullable(field.data_type)

                # unpack list or map
                while is_list_type(data_type) or is_map_type(data_type):
                    data_type = (data_type.value_data_type if
                        is_map_type(data_type) else data_type.data_type)

                if is_user_defined_type(data_type):
                    import_classes.append(fmt_class_prefix(data_type))

        import_classes = list(set(import_classes))
        import_classes.sort()

        return import_classes

    def _generate_imports_h(self, import_classes):
        import_classes = list(set(import_classes))
        import_classes.sort()

        for import_class in import_classes:
            self.emit('@class {};'.format(import_class))

        if import_classes:
            self.emit()

    def _generate_imports_m(self, import_classes):
        import_classes = list(set(import_classes))
        import_classes.sort()

        for import_class in import_classes:
            self.emit(fmt_import(import_class))

        self.emit()

    def _generate_init_imports_h(self, data_type):
        self.emit('#import <Foundation/Foundation.h>')
        self.emit()
        self.emit('#import "DBSerializableProtocol.h"')

        if data_type.parent_type and not is_union_type(data_type):
            self.emit(fmt_import(fmt_class_prefix(data_type.parent_type)))

        self.emit()

    def _get_namespace_route_imports(self,
                                     namespace,
                                     include_route_args=True,
                                     include_route_deep_args=False):
        result = []

        def _unpack_and_store_data_type(data_type):
            data_type, _ = unwrap_nullable(data_type)
            if is_list_type(data_type):
                while is_list_type(data_type):
                    data_type, _ = unwrap_nullable(data_type.data_type)

            if not is_void_type(data_type) and is_user_defined_type(data_type):
                result.append(data_type)

        for route in namespace.routes:
            if include_route_args:
                data_type, _ = unwrap_nullable(route.arg_data_type)
                _unpack_and_store_data_type(data_type)
            elif include_route_deep_args:
                data_type, _ = unwrap_nullable(route.arg_data_type)
                if is_union_type(data_type) or is_list_type(data_type):
                    _unpack_and_store_data_type(data_type)
                elif not is_void_type(data_type):
                    for field in data_type.all_fields:
                        data_type, _ = unwrap_nullable(field.data_type)
                        if (is_struct_type(data_type) or
                                is_union_type(data_type) or
                                is_list_type(data_type)):
                            _unpack_and_store_data_type(data_type)

            _unpack_and_store_data_type(route.result_data_type)
            _unpack_and_store_data_type(route.error_data_type)

        return result

    def _cstor_name_from_fields(self, fields):
        """Returns an Obj C appropriate name for a constructor based on
        the name of the first argument."""
        if fields:
            return self._cstor_name_from_field(fields[0])
        else:
            return 'initDefault'

    def _cstor_name_from_field(self, field):
        """Returns an Obj C appropriate name for a constructor based on
        the name of the supplied argument."""
        return 'initWith{}'.format(fmt_camel_upper(field.name))

    def _cstor_name_from_fields_names(self, fields_names):
        """Returns an Obj C appropriate name for a constructor based on
        the name of the first argument."""
        if fields_names:
            return 'initWith{}'.format(fmt_camel_upper(fields_names[0][0]))
        else:
            return 'initDefault'

    def _struct_has_defaults(self, struct):
        """Returns whether the given struct has any default values."""
        return [
            f for f in struct.all_fields
            if f.has_default or is_nullable_type(f.data_type)
        ]



#!/usr/bin/env python

import textwrap
import unittest

from stone.frontend.frontend import specs_to_ir


class TestStone(unittest.TestCase):
    def _compare_namespace_names(self, api, expected_names):
        expected_names = sorted(expected_names)
        actual_names = sorted(list(api.namespaces.keys()))
        self.assertEqual(expected_names, actual_names)

    def _compare_datatype_names(self, namespace, expected_names):
        expected_names = sorted(expected_names)
        actual_names = sorted(list(namespace.data_type_by_name.keys()))
        self.assertEqual(expected_names, actual_names)
        actual_names_on_datatypes = sorted([d.name for d in namespace.data_type_by_name.values()])
        self.assertEqual(actual_names, actual_names_on_datatypes)

    def test_simple(self):
        """
        Tests that route whitelisting can generate the right datatypes for a namespace.
        """
        text = textwrap.dedent("""\
            namespace test

            struct TestArg
                f String
                    "test doc"
                example default
                    f = "asdf"
            struct TestResult
                f String
                    "test doc"
                example default
                    f = "asdf"
            route TestRoute (TestArg, TestResult, Void)
                "test doc"

            struct TestArg2
                f String
                    "test doc"
                example default
                    f = "asdf"
            struct TestResult2
                f String
                    "test doc"
                example default
                    f = "asdf"
            route TestRoute:2 (TestArg2, TestResult2, Void)
                "test doc"

            """)
        route_whitelist_filter = {
            "route_whitelist": {"test": ["TestRoute:2"]},
            "datatype_whitelist": {}
        }
        api = specs_to_ir([('test.stone', text)], route_whitelist_filter=route_whitelist_filter)
        self._compare_namespace_names(api, ['test'])
        self._compare_datatype_names(api.namespaces['test'], ['TestArg2', 'TestResult2'])

    def test_stable_ordering(self):
        """
        Tests that route generation returns the same route order on generation.
        """
        text = textwrap.dedent("""\
            namespace test

            struct TestArg
                f String
                    "test doc"
                example default
                    f = "asdf"
            struct TestResult
                f String
                    "test doc"
                example default
                    f = "asdf"
            route TestRoute (TestArg, TestResult, Void)
                "test doc"

            route OtherRoute (Void, Void, Void)
                "Additional Route"

            """)

        route_whitelist = ["OtherRoute", "TestRoute"]

        for x in range(2, 100):
            text += textwrap.dedent("""\
            route TestRoute:{} (TestArg, TestResult, Void)
                "test doc"
            """.format(x))
            route_whitelist.append("TestRoute:{}".format(x))

        route_whitelist_filter = {
            "route_whitelist": {"test": route_whitelist},
            "datatype_whitelist": {}
        }

        api = specs_to_ir([('test.stone', text)], route_whitelist_filter=route_whitelist_filter)
        routes = api.namespaces['test'].routes
        self.assertEqual(len(routes), 100)
        self.assertEqual(routes[0].name, "OtherRoute")

        for x in range(1, 100):
            self.assertEqual(routes[x].name, "TestRoute")
            self.assertEqual(routes[x].version, x)

    def test_star(self):
        """
        Tests that inputs with "*" work as expected.
        """
        text = textwrap.dedent("""\
            namespace test

            struct TestArg
                f String
                    "test doc"
                example default
                    f = "asdf"
            struct TestResult
                f String
                    "test doc"
                example default
                    f = "asdf"
            route TestRoute (TestArg, TestResult, Void)
                "test doc"

            struct TestArg2
                f String
                    "test doc"
                example default
                    f = "asdf"
            struct TestResult2
                f String
                    "test doc"
                example default
                    f = "asdf"
            route TestRoute:2 (TestArg2, TestResult2, Void)
                "test doc"

            """)
        route_whitelist_filter = {
            "route_whitelist": {"test": ["*"]},
            "datatype_whitelist": {}
        }
        api = specs_to_ir([('test.stone', text)],
                          route_whitelist_filter=route_whitelist_filter)
        self._compare_namespace_names(api, ['test'])
        self._compare_datatype_names(api.namespaces['test'],
                                     ['TestArg', 'TestArg2', 'TestResult', 'TestResult2'])

    def test_alias(self):
        """
        Tests that aliased datatypes are correctly generated.
        """
        text = textwrap.dedent("""\
            namespace test

            struct TestArg
                f String
                    "test doc"
                example default
                    f = "asdf"
            struct TestResult
                f String
                    "test doc"
                example default
                    f = "asdf"
            alias TestAlias = TestArg
            route TestRoute (TestAlias, TestResult, Void)
                "test doc"

            """)
        route_whitelist_filter = {
            "route_whitelist": {"test": ["TestRoute"]},
            "datatype_whitelist": {}
        }
        api = specs_to_ir([('test.stone', text)], route_whitelist_filter=route_whitelist_filter)
        self._compare_namespace_names(api, ['test'])
        self._compare_datatype_names(api.namespaces['test'], ['TestArg', 'TestResult'])

    def test_imports(self):
        """
        Tests that datatypes imported from another namespace are correctly included.
        """
        text = textwrap.dedent("""\
            namespace test

            struct TestArg
                f String
                    "test doc"
                example default
                    f = "asdf"
            struct TestResult
                f String
                    "test doc"
                example default
                    f = "asdf"

            """)
        text2 = textwrap.dedent("""\
            namespace test2

            import test

            route TestRoute (test.TestArg, test.TestResult, Void)
                "test doc"

            """)
        route_whitelist_filter = {
            "route_whitelist": {"test2": ["TestRoute"]},
            "datatype_whitelist": {}
        }
        api = specs_to_ir([('test.stone', text), ('test2.stone', text2)],
                          route_whitelist_filter=route_whitelist_filter)
        self._compare_namespace_names(api, ['test', 'test2'])
        self._compare_datatype_names(api.namespaces['test'], ['TestArg', 'TestResult'])
        self._compare_datatype_names(api.namespaces['test2'], [])

    def test_builtin_types(self):
        """
        Tests that builtin datatypes, like lists, maps, and unions, are correctly evaluated.
        """
        text = textwrap.dedent("""\
            namespace test

            union Foo
                a
                    "test doc"
                example default
                    a = null
            union Bar
                a
                    "test doc"
                example default
                    a = null
            struct TestArg
                f String
                    "test doc"
                example default
                    f = "asdf"
            struct TestResult
                f List(Foo)
                    "test doc"
                b Map(String, Bar)
                    "test doc"
                example default
                    f = [default]
                    b = {"test": default}
            route TestRoute (TestArg, TestResult, Void)
                "test doc"

            """)
        route_whitelist_filter = {
            "route_whitelist": {"test": ["TestRoute"]},
            "datatype_whitelist": {}
        }
        api = specs_to_ir([('test.stone', text)], route_whitelist_filter=route_whitelist_filter)
        self._compare_namespace_names(api, ['test'])
        self._compare_datatype_names(api.namespaces['test'], ['TestArg', 'TestResult',
                                                              'Foo', 'Bar'])

    def test_subtype(self):
        """
        Tests that datatypes that inherit from others are correctly generated.
        """
        text = textwrap.dedent("""\
            namespace test

            union Foo
                a
                    "test doc"
                example default
                    a = null
            union Bar extends Foo
                b
                    "test doc"
                example default
                    a = null
            struct TestArg
                f String
                    "test doc"
                example default
                    f = "asdf"
            struct TestResult
                f Bar
                    "test doc"
                example default
                    f = default
            route TestRoute (TestArg, TestResult, Void)
                "test doc"

            """)
        route_whitelist_filter = {
            "route_whitelist": {"test": ["TestRoute"]},
            "datatype_whitelist": {}
        }
        api = specs_to_ir([('test.stone', text)], route_whitelist_filter=route_whitelist_filter)
        self._compare_namespace_names(api, ['test'])
        self._compare_datatype_names(api.namespaces['test'], ['TestArg', 'TestResult',
                                                              'Foo', 'Bar'])

        # Test enumerated subtypes as well
        text = textwrap.dedent("""\
            namespace test

            struct Foo
                union
                    file File
                    folder Folder
                path String
                    "test doc"
                example default
                    file = default
                example folder_default
                    folder = default

            struct File extends Foo
                a String
                    "test doc"
                example default
                    a = "a"
                    path = "a"

            struct Folder extends Foo
                b String
                    "test doc"
                example default
                    b = "b"
                    path = "a"

            struct TestArg
                f Foo
                    "test doc"
                example default
                    f = default
            struct TestResult
                f String
                    "test doc"
                example default
                    f = "asdf"
            route TestRoute (TestArg, TestResult, Void)
                "test doc"

            """)
        route_whitelist_filter = {
            "route_whitelist": {"test": ["TestRoute"]},
            "datatype_whitelist": {}
        }
        api = specs_to_ir([('test.stone', text)], route_whitelist_filter=route_whitelist_filter)
        self._compare_namespace_names(api, ['test'])
        self._compare_datatype_names(api.namespaces['test'], ['TestArg', 'TestResult',
                                                              'Foo', 'File', 'Folder'])

    def test_doc_refs(self):
        """
        Tests that datatypes referenced in documentation get generated.
        """
        text = textwrap.dedent("""\
            namespace test

            union Foo
                a
                    "test doc"
                example default
                    a = null
            union Bar
                a
                    "test doc"
                example default
                    a = null

            """)

        text2 = textwrap.dedent("""\
            namespace test2
                ":type:`test.Foo`"

            alias TestAlias = String
                ":field:`Baz.a`"

            struct TestStruct
                a TestAlias
                    "test doc"
                example default
                    a = "asdf"

            struct Baz
                a String
                    "test doc :field:`TestStruct.a`"
                example default
                    a = "asdf"

            struct TestArg
                f String
                    "test doc"
                example default
                    f = "asdf"
            struct TestResult
                f String
                    "test doc"
                example default
                    f = "asdf"
            route TestRoute (TestArg, TestResult, Void)
                ":type:`Baz` test doc"
            """)
        route_whitelist_filter = {
            "route_whitelist": {"test2": ["TestRoute"]},
            "datatype_whitelist": {}
        }
        api = specs_to_ir([('test.stone', text), ('test2.stone', text2)],
                          route_whitelist_filter=route_whitelist_filter)
        self._compare_namespace_names(api, ['test', 'test2'])
        self._compare_datatype_names(api.namespaces['test'], ['Foo'])
        self._compare_datatype_names(api.namespaces['test2'], ['TestArg', 'TestResult',
                                                               'TestStruct', 'Baz'])

if __name__ == '__main__':
    unittest.main()



"""Example backend that outputs a Stone file equivalent to the input file.

Current limitations:
- Whitespace is not reproduced exactly (this may be a feature)
- Order of definitions is lost
- Comments are lost
- Aliases are lost (they are expanded in-line)
- Docstrings are reformatted
"""

from stone.backend import CodeBackend
from stone.frontend.ast import AstTypeRef
from stone.ir import DataType
from stone.ir import List, String, Timestamp
from stone.ir import Struct, Union, Void
from stone.ir.data_types import _BoundedInteger, _BoundedFloat


class UnstoneBackend(CodeBackend):
    """Main class.

    The Stone CLI finds this class through introspection."""

    def generate(self, api):
        """Main code generator entry point."""
        # Create a file for each namespace.
        for namespace in api.namespaces.values():
            with self.output_to_relative_path('%s.stone' % namespace.name):
                # Output a namespace header.
                self.emit('namespace %s' % namespace.name)
                # Output all data type (struct and union) definitions.
                for data_type in namespace.linearize_data_types():
                    self.generate_data_type(data_type)
                # Output all route definitions.
                for route in namespace.routes:
                    self.generate_route(route)

    def generate_data_type(self, data_type):
        """Output a data type definition (a struct or union)."""
        if isinstance(data_type, Struct):
            # Output a struct definition.
            self.emit('')
            self.emit('struct %s' % data_type.name)
            with self.indent():
                if data_type.doc is not None:
                    self.emit(self.format_string(data_type.doc))
                for field in data_type.fields:
                    type_repr = self.format_data_type(field.data_type)
                    if not field.has_default:
                        self.emit('{} {}'.format(field.name, type_repr))
                    else:
                        self.emit('%s %s = %s' %
                            (field.name, type_repr, self.format_value(field.default)))
                    if field.doc is not None:
                        with self.indent():
                            self.emit(self.format_value(field.doc))
        elif isinstance(data_type, Union):
            # Output a union definition.
            self.emit('')
            self.emit('union %s' % data_type.name)
            with self.indent():
                if data_type.doc is not None:
                    self.emit(self.format_string(data_type.doc))
                for field in data_type.fields:
                    name = field.name
                    # Add a star for a catch-all field.
                    # (There are two ways to recognize these.)
                    if field.catch_all or field is data_type.catch_all_field:
                        name += '*'
                    if isinstance(field.data_type, Void):
                        self.emit('%s' % (name))
                    else:
                        type_repr = self.format_data_type(field.data_type)
                        self.emit('{} {}'.format(name, type_repr))
                    if field.doc is not None:
                        with self.indent():
                            self.emit(self.format_value(field.doc))
        else:
            # Don't know what this is.
            self.emit('')
            self.emit('# ??? %s' % repr(data_type))

    def generate_route(self, route):
        """Output a route definition."""
        self.emit('')
        self.emit('route {} ({}, {}, {})'.format(
            route.name,
            self.format_data_type(route.arg_data_type),
            self.format_data_type(route.result_data_type),
            self.format_data_type(route.error_data_type)
        ))
        # Output the docstring.
        with self.indent():
            if route.doc is not None:
                self.emit(self.format_string(route.doc))

    # Table describing data types with parameters.
    _data_type_map = [
        (_BoundedInteger, ['min_value', 'max_value']),
        (_BoundedFloat, ['min_value', 'max_value']),
        (List, ['data_type', 'min_items', 'max_items']),
        (String, ['min_length', 'max_length', 'pattern']),
        (Timestamp, ['format']),
    ]

    def format_data_type(self, data_type):
        """Helper function to format a data type.

        This returns the name if it's a struct or union, otherwise
        (i.e. for primitive types) it renders the name and the
        parameters.
        """
        s = data_type.name
        for type_class, key_list in self._data_type_map:
            if isinstance(data_type, type_class):
                args = []
                for key in key_list:
                    val = getattr(data_type, key)
                    if val is not None:
                        if isinstance(val, AstTypeRef):
                            sval = val.name
                        elif isinstance(val, DataType):
                            sval = self.format_data_type(val)
                        else:
                            sval = self.format_value(val)
                        args.append(key + '=' + sval)
                if args:
                    s += '(' + ', '.join(args) + ')'
                break
        if data_type.nullable:
            s += '?'
        return s

    def format_value(self, val):
        """Helper function to format a value."""
        if isinstance(val, str):
            return self.format_string(val)
        else:
            return str(val)

    def format_string(self, val):
        """Helper function to format a string."""
        return '"' + val.replace('\\', '\\\\').replace('"', '\\"').replace('\n', '\n\n') + '"'



import textwrap

from stone.backends.python_client import PythonClientBackend
from stone.ir import (
    ApiNamespace,
    ApiRoute,
    Int32,
    List,
    Map,
    Nullable,
    String,
    Void,
    StructField,
    Struct,
)

MYPY = False
if MYPY:
    import typing  # noqa: F401 # pylint: disable=import-error,unused-import,useless-suppression

import unittest

class TestGeneratedPythonClient(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _evaluate_namespace(self, ns):
        # type: (ApiNamespace) -> typing.Text

        backend = PythonClientBackend(
            target_folder_path='output',
            args=['-a', 'scope', '-a', 'another_attribute',
                  '-m', 'files', '-c', 'DropboxBase', '-t', 'dropbox'])
        backend._generate_routes(ns)
        return backend.output_buffer_to_string()

    def _evaluate_namespace_with_auth_mode(self, ns, auth_mode):
        # type: (ApiNamespace, str) -> typing.Text

        # supply supported auth modes to the SDK generator using the new syntax
        backend = PythonClientBackend(
            target_folder_path='output',
            args=['-w', auth_mode, '-m', 'files', '-c', 'DropboxBase', '-t', 'dropbox'])
        backend._generate_route_methods({ns})
        return backend.output_buffer_to_string()

    def test_route_with_version_number(self):
        # type: () -> None

        route1 = ApiRoute('get_metadata', 1, None)
        route1.set_attributes(None, ':route:`get_metadata:2`', Void(), Void(), Void(), {})
        route2 = ApiRoute('get_metadata', 2, None)
        route2.set_attributes(None, None, Void(), Int32(), Void(), {})
        ns = ApiNamespace('files')
        ns.add_route(route1)
        ns.add_route(route2)

        result = self._evaluate_namespace(ns)

        expected = textwrap.dedent('''\
            def files_get_metadata(self):
                """
                :meth:`files_get_metadata_v2`

                :rtype: None
                """
                arg = None
                r = self.request(
                    files.get_metadata,
                    'files',
                    arg,
                    None,
                )
                return None

            def files_get_metadata_v2(self):
                arg = None
                r = self.request(
                    files.get_metadata_v2,
                    'files',
                    arg,
                    None,
                )
                return r

        ''')

        self.assertEqual(result, expected)

    def test_route_with_auth_mode1(self):
        # type: () -> None

        route1 = ApiRoute('get_metadata', 1, None)
        route1.set_attributes(None, ':route:`get_metadata:2`', Void(), Void(), Void(),
                              {'auth': 'app'})
        route2 = ApiRoute('get_metadata', 2, None)
        route2.set_attributes(None, None, Void(), Int32(), Void(),
                              {'auth': 'user, app'})
        ns = ApiNamespace('files')
        ns.add_route(route1)
        ns.add_route(route2)

        result = self._evaluate_namespace_with_auth_mode(ns, 'user')

        expected = textwrap.dedent('''\
            # ------------------------------------------
            # Routes in files namespace

            def files_get_metadata_v2(self):
                arg = None
                r = self.request(
                    files.get_metadata_v2,
                    'files',
                    arg,
                    None,
                )
                return r

        ''')

        self.assertEqual(result, expected)

    def test_route_with_auth_mode2(self):
        # type: () -> None

        route1 = ApiRoute('get_metadata', 1, None)
        route1.set_attributes(None, ':route:`get_metadata:2`', Void(), Void(), Void(),
                              {'auth': 'user'})
        route2 = ApiRoute('get_metadata', 2, None)
        route2.set_attributes(None, None, Void(), Int32(), Void(),
                              {'auth': 'user, app'})
        ns = ApiNamespace('files')
        ns.add_route(route1)
        ns.add_route(route2)

        result = self._evaluate_namespace_with_auth_mode(ns, 'user')

        expected = textwrap.dedent('''\
            # ------------------------------------------
            # Routes in files namespace

            def files_get_metadata(self):
                """
                :meth:`files_get_metadata_v2`

                :rtype: None
                """
                arg = None
                r = self.request(
                    files.get_metadata,
                    'files',
                    arg,
                    None,
                )
                return None

            def files_get_metadata_v2(self):
                arg = None
                r = self.request(
                    files.get_metadata_v2,
                    'files',
                    arg,
                    None,
                )
                return r

        ''')

        self.assertEqual(result, expected)

    def test_route_with_auth_mode3(self):
        # type: () -> None

        route1 = ApiRoute('get_metadata', 1, None)
        route1.set_attributes(None, ':route:`get_metadata:2`', Void(), Void(), Void(),
                             {'auth': 'app'})
        route2 = ApiRoute('get_metadata', 2, None)
        route2.set_attributes(None, None, Void(), Int32(), Void(),
                             {'auth': 'app, team'})
        ns = ApiNamespace('files')
        ns.add_route(route1)
        ns.add_route(route2)

        result = self._evaluate_namespace_with_auth_mode(ns, 'user')

        expected = textwrap.dedent('''\
            # ------------------------------------------
            # Routes in files namespace

        ''')

        self.assertEqual(result, expected)

    def test_route_with_version_number_name_conflict(self):
        # type: () -> None

        route1 = ApiRoute('get_metadata', 2, None)
        route1.set_attributes(None, None, Void(), Int32(), Void(), {})
        route2 = ApiRoute('get_metadata_v2', 1, None)
        route2.set_attributes(None, None, Void(), Void(), Void(), {})
        ns = ApiNamespace('files')
        ns.add_route(route1)
        ns.add_route(route2)

        with self.assertRaises(RuntimeError) as cm:
            self._evaluate_namespace(ns)
        self.assertEqual(
            'There is a name conflict between {!r} and {!r}'.format(route1, route2),
            str(cm.exception))

    def test_route_argument_doc_string(self):
        backend = PythonClientBackend(
            target_folder_path='output',
            args=['-m', 'files', '-c', 'DropboxBase', '-t', 'dropbox'])
        ns = ApiNamespace('files')
        self.assertEqual(backend._format_type_in_doc(ns, Int32()), 'int')
        self.assertEqual(backend._format_type_in_doc(ns, Void()), 'None')
        self.assertEqual(backend._format_type_in_doc(ns, List(String())), 'List[str]')
        self.assertEqual(backend._format_type_in_doc(ns, Nullable(String())),
                         'Nullable[str]')
        self.assertEqual(backend._format_type_in_doc(ns, Map(String(), Int32())),
                         'Map[str, int]')

    def test_route_with_attributes_in_docstring(self):
        # type: () -> None

        route = ApiRoute('get_metadata', 1, None)
        route.set_attributes(None, None, Void(), Void(), Void(), {
            'scope': 'events.read', 'another_attribute': 'foo'
        })
        ns = ApiNamespace('files')
        ns.add_route(route)

        result = self._evaluate_namespace(ns)
        expected = textwrap.dedent('''\
            def files_get_metadata(self):
                """
                Route attributes:
                    scope: events.read
                    another_attribute: foo

                :rtype: None
                """
                arg = None
                r = self.request(
                    files.get_metadata,
                    'files',
                    arg,
                    None,
                )
                return None

        ''')
        self.assertEqual(result, expected)

    def test_route_with_none_attribute_in_docstring(self):
        # type: () -> None

        route = ApiRoute('get_metadata', 1, None)
        route.set_attributes(None, None, Void(), Void(), Void(), {
            'scope': 'events.read', 'another_attribute': None
        })
        ns = ApiNamespace('files')
        ns.add_route(route)

        result = self._evaluate_namespace(ns)
        expected = textwrap.dedent('''\
            def files_get_metadata(self):
                """
                Route attributes:
                    scope: events.read

                :rtype: None
                """
                arg = None
                r = self.request(
                    files.get_metadata,
                    'files',
                    arg,
                    None,
                )
                return None

        ''')
        self.assertEqual(result, expected)

    def test_route_with_attributes_and_doc_in_docstring(self):
        # type: () -> None
        """
        In particular make sure there's spacing b/w overview and attrs.
        """

        route = ApiRoute('get_metadata', 1, None)
        route.set_attributes(None, "Test string.", Void(), Void(), Void(),
                             {'scope': 'events.read'})
        ns = ApiNamespace('files')
        ns.add_route(route)

        result = self._evaluate_namespace(ns)
        expected = textwrap.dedent('''\
            def files_get_metadata(self):
                """
                Test string.

                Route attributes:
                    scope: events.read

                :rtype: None
                """
                arg = None
                r = self.request(
                    files.get_metadata,
                    'files',
                    arg,
                    None,
                )
                return None

        ''')
        self.assertEqual(result, expected)

    def test_route_with_doc_and_attribute_and_data_types(self):
        # type: () -> None
        ns = ApiNamespace('files')
        struct = Struct('MyStruct', ns, None)
        struct.set_attributes(None, [
            StructField('field1', Int32(), None, None),
            StructField('field2', Int32(), None, None),
        ])

        route = ApiRoute('test/route', 1, None)
        route.set_attributes(
            None, "Test string.", struct, Int32(), Void(), {'scope': 'events.read'}
        )
        ns.add_route(route)

        result = self._evaluate_namespace(ns)
        expected = textwrap.dedent('''\
            def files_test_route(self,
                                 field1,
                                 field2):
                """
                Test string.

                Route attributes:
                    scope: events.read

                :type field1: int
                :type field2: int
                :rtype: int
                """
                arg = files.MyStruct(field1,
                                     field2)
                r = self.request(
                    files.test_route,
                    'files',
                    arg,
                    None,
                )
                return r

        ''')
        self.assertEqual(result, expected)

    # TODO: add more unit tests for client code generation



import json
import sys

from stone.backend import CodeBackend
from stone.backends.js_helpers import (
    fmt_jsdoc_union,
    fmt_type,
    fmt_type_name,
)
from stone.ir import (
    is_user_defined_type,
    is_union_type,
    is_struct_type,
    is_void_type,
    unwrap,
)

_MYPY = False
if _MYPY:
    import typing  # noqa: F401 # pylint: disable=import-error,unused-import,useless-suppression

import argparse

_cmdline_parser = argparse.ArgumentParser(prog='js-types-backend')
_cmdline_parser.add_argument(
    'filename',
    help=('The name to give the single Javascript file that is created and '
          'contains all of the JSDoc types.'),
)
_cmdline_parser.add_argument(
    '-e',
    '--extra-arg',
    action='append',
    type=str,
    default=[],
    help=("Additional properties to add to a route's argument type based "
          "on if the route has a certain attribute set. Format (JSON): "
          '{"match": ["ROUTE_ATTR", ROUTE_VALUE_TO_MATCH], '
          '"arg_name": "ARG_NAME", "arg_type": "ARG_TYPE", '
          '"arg_docstring": "ARG_DOCSTRING"}'),
)

_header = """\
// Auto-generated by Stone, do not modify.
/**
 * An Error object returned from a route.
 * @typedef {Object} Error
 * @property {string} error_summary - Text summary of the error.
 * @property {T} error - The error object.
 * @property {UserMessage} user_message - An optional field. If present, it includes a
    message that can be shown directly to the end user of your app. You should show this message
    if your app is unprepared to programmatically handle the error returned by an endpoint.
 * @template T
 */

/**
 * User-friendly error message.
 * @typedef {Object} UserMessage
 * @property {string} text - The message.
 * @property {string} locale
 */

 /**
  * @typedef {string} Timestamp
  */
"""


class JavascriptTypesBackend(CodeBackend):
    """Generates a single Javascript file with all of the data types defined in JSDoc."""

    cmdline_parser = _cmdline_parser

    preserve_aliases = True

    def generate(self, api):
        with self.output_to_relative_path(self.args.filename):

            self.emit_raw(_header)

            extra_args = self._parse_extra_args(api, self.args.extra_arg)

            for namespace in api.namespaces.values():
                for data_type in namespace.data_types:
                    self._generate_type(data_type, extra_args.get(data_type, []))

    def _parse_extra_args(self, api, extra_args_raw):
        """
        Parses extra arguments into a map keyed on particular data types.
        """
        extra_args = {}

        def die(m, extra_arg_raw):
            print('Invalid --extra-arg:{}: {}'.format(m, extra_arg_raw),
                  file=sys.stderr)
            sys.exit(1)

        for extra_arg_raw in extra_args_raw:
            try:
                extra_arg = json.loads(extra_arg_raw)
            except ValueError as e:
                die(str(e), extra_arg_raw)

            # Validate extra_arg JSON blob
            if 'match' not in extra_arg:
                die('No match key', extra_arg_raw)
            elif (not isinstance(extra_arg['match'], list) or
                    len(extra_arg['match']) != 2):
                die('match key is not a list of two strings', extra_arg_raw)
            elif (not isinstance(extra_arg['match'][0], str) or
                    not isinstance(extra_arg['match'][1], str)):
                print(type(extra_arg['match'][0]))
                die('match values are not strings', extra_arg_raw)
            elif 'arg_name' not in extra_arg:
                die('No arg_name key', extra_arg_raw)
            elif not isinstance(extra_arg['arg_name'], str):
                die('arg_name is not a string', extra_arg_raw)
            elif 'arg_type' not in extra_arg:
                die('No arg_type key', extra_arg_raw)
            elif not isinstance(extra_arg['arg_type'], str):
                die('arg_type is not a string', extra_arg_raw)
            elif ('arg_docstring' in extra_arg and
                    not isinstance(extra_arg['arg_docstring'], str)):
                die('arg_docstring is not a string', extra_arg_raw)

            attr_key, attr_val = extra_arg['match'][0], extra_arg['match'][1]

            extra_args.setdefault(attr_key, {})[attr_val] = \
                (extra_arg['arg_name'], extra_arg['arg_type'],
                 extra_arg.get('arg_docstring'))

        # Extra arguments, keyed on data type objects.
        extra_args_for_types = {}
        # Locate data types that contain extra arguments
        for namespace in api.namespaces.values():
            for route in namespace.routes:
                extra_parameters = []
                if is_user_defined_type(route.arg_data_type):
                    for attr_key in route.attrs:
                        if attr_key not in extra_args:
                            continue
                        attr_val = route.attrs[attr_key]
                        if attr_val in extra_args[attr_key]:
                            extra_parameters.append(extra_args[attr_key][attr_val])
                if len(extra_parameters) > 0:
                    extra_args_for_types[route.arg_data_type] = extra_parameters

        return extra_args_for_types

    def _generate_type(self, data_type, extra_parameters):
        if is_struct_type(data_type):
            self._generate_struct(data_type, extra_parameters)
        elif is_union_type(data_type):
            self._generate_union(data_type)

    def _emit_jsdoc_header(self, doc=None):
        self.emit()
        self.emit('/**')
        if doc:
            self.emit_wrapped_text(self.process_doc(doc, self._docf), prefix=' * ')

    def _generate_struct(self, struct_type, extra_parameters=None, nameOverride=None):
        """
        Emits a JSDoc @typedef for a struct.
        """
        extra_parameters = extra_parameters if extra_parameters is not None else []
        self._emit_jsdoc_header(struct_type.doc)
        self.emit(
            ' * @typedef {Object} %s' % (
                nameOverride if nameOverride else fmt_type_name(struct_type)
            )
        )

        # Some structs can explicitly list their subtypes. These structs
        # have a .tag field that indicate which subtype they are.
        if struct_type.is_member_of_enumerated_subtypes_tree():
            if struct_type.has_enumerated_subtypes():
                # This struct is the parent to multiple subtypes.
                # Determine all of the possible values of the .tag
                # property.
                tag_values = []
                for tags, _ in struct_type.get_all_subtypes_with_tags():
                    for tag in tags:
                        tag_values.append('"%s"' % tag)

                jsdoc_tag_union = fmt_jsdoc_union(tag_values)
                txt = '@property {%s} .tag - Tag identifying the subtype variant.' % \
                    jsdoc_tag_union
                self.emit_wrapped_text(txt)
            else:
                # This struct is a particular subtype. Find the applicable
                # .tag value from the parent type, which may be an
                # arbitrary number of steps up the inheritance hierarchy.
                parent = struct_type.parent_type
                while not parent.has_enumerated_subtypes():
                    parent = parent.parent_type
                # parent now contains the closest parent type in the
                # inheritance hierarchy that has enumerated subtypes.
                # Determine which subtype this is.
                for subtype in parent.get_enumerated_subtypes():
                    if subtype.data_type == struct_type:
                        txt = '@property {\'%s\'} [.tag] - Tag identifying ' \
                            'this subtype variant. This field is only ' \
                            'present when needed to discriminate ' \
                            'between multiple possible subtypes.' % \
                            subtype.name
                        self.emit_wrapped_text(txt)
                        break

        for param_name, param_type, param_docstring in extra_parameters:
            param_docstring = ' - %s' % param_docstring if param_docstring else ''
            self.emit_wrapped_text(
                '@property {{{}}} [{}]{}'.format(
                    param_type,
                    param_name,
                    param_docstring,
                ),
                prefix=' * ',
            )

        # NOTE: JSDoc @typedef does not support inheritance. Using @class would be inappropriate,
        # since these are not nominal types backed by a constructor. Thus, we emit all_fields,
        # which includes fields on parent types.
        for field in struct_type.all_fields:
            field_doc = ' - ' + field.doc if field.doc else ''
            field_type, nullable, _ = unwrap(field.data_type)
            field_js_type = fmt_type(field_type)
            # Translate nullable types into optional properties.
            field_name = '[' + field.name + ']' if nullable else field.name
            self.emit_wrapped_text(
                '@property {{{}}} {}{}'.format(
                    field_js_type,
                    field_name,
                    self.process_doc(field_doc, self._docf),
                ),
                prefix=' * ',
            )

        self.emit(' */')

    def _generate_union(self, union_type):
        """
        Emits a JSDoc @typedef for a union type.
        """
        union_name = fmt_type_name(union_type)
        self._emit_jsdoc_header(union_type.doc)
        self.emit(' * @typedef {Object} %s' % union_name)
        variant_types = []
        for variant in union_type.all_fields:
            variant_types.append("'%s'" % variant.name)
            variant_data_type, _, _ = unwrap(variant.data_type)
            # Don't emit fields for void types.
            if not is_void_type(variant_data_type):
                variant_doc = ' - Available if .tag is %s.' % variant.name
                if variant.doc:
                    variant_doc += ' ' + variant.doc
                self.emit_wrapped_text(
                    '@property {{{}}} [{}]{}'.format(
                        fmt_type(variant_data_type),
                        variant.name,
                        variant_doc,
                    ),
                    prefix=' * ',
                )
        jsdoc_tag_union = fmt_jsdoc_union(variant_types)
        self.emit(' * @property {%s} .tag - Tag identifying the union variant.' % jsdoc_tag_union)
        self.emit(' */')

    def _docf(self, tag, val):  # pylint: disable=unused-argument
        """
        Callback used as the handler argument to process_docs(). This converts
        Stone doc references to JSDoc-friendly annotations.
        """
        # TODO(kelkabany): We're currently just dropping all doc ref tags.
        return val



from collections import OrderedDict


class ASTNode:

    def __init__(self, path, lineno, lexpos):
        """
        Args:
            lineno (int): The line number where the start of this element
                occurs.
            lexpos (int): The character offset into the file where this element
                occurs.
        """
        self.path = path
        self.lineno = lineno
        self.lexpos = lexpos


class AstNamespace(ASTNode):

    def __init__(self, path, lineno, lexpos, name, doc):
        """
        Args:
            name (str): The namespace of the spec.
            doc (Optional[str]): The docstring for this namespace.
        """
        super().__init__(path, lineno, lexpos)
        self.name = name
        self.doc = doc

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return 'AstNamespace({!r})'.format(self.name)


class AstImport(ASTNode):

    def __init__(self, path, lineno, lexpos, target):
        """
        Args:
            target (str): The name of the namespace to import.
        """
        super().__init__(path, lineno, lexpos)
        self.target = target

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return 'AstImport({!r})'.format(self.target)

class AstAlias(ASTNode):

    def __init__(self, path, lineno, lexpos, name, type_ref, doc):
        """
        Args:
            name (str): The name of the alias.
            type_ref (AstTypeRef): The data type of the field.
            doc (Optional[str]): Documentation string for the alias.
        """
        super().__init__(path, lineno, lexpos)
        self.name = name
        self.type_ref = type_ref
        self.doc = doc
        self.annotations = []

    def set_annotations(self, annotations):
        self.annotations = annotations

    def __repr__(self):
        return 'AstAlias({!r}, {!r})'.format(self.name, self.type_ref)

class AstTypeDef(ASTNode):

    def __init__(self, path, lineno, lexpos, name, extends, doc, fields,
                 examples):
        """
        Args:
            name (str): Name assigned to the type.
            extends (Optional[str]); Name of the type this inherits from.
            doc (Optional[str]): Docstring for the type.
            fields (List[AstField]): Fields of a type, not including
                inherited ones.
            examples (Optional[OrderedDict[str, AstExample]]): Map from label
                to example.
        """
        super().__init__(path, lineno, lexpos)

        self.name = name
        assert isinstance(extends, (AstTypeRef, type(None))), type(extends)
        self.extends = extends
        assert isinstance(doc, (str, type(None)))
        self.doc = doc
        assert isinstance(fields, list)
        self.fields = fields
        assert isinstance(examples, (OrderedDict, type(None))), type(examples)
        self.examples = examples

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return 'AstTypeDef({!r}, {!r}, {!r})'.format(
            self.name,
            self.extends,
            self.fields,
        )

class AstStructDef(AstTypeDef):

    def __init__(self, path, lineno, lexpos, name, extends, doc, fields,
                 examples, subtypes=None):
        """
        Args:
            subtypes (Tuple[List[AstSubtypeField], bool]): Inner list
                enumerates subtypes. The bool indicates whether this struct
                is a catch-all.

        See AstTypeDef for other constructor args.
        """

        super().__init__(
            path, lineno, lexpos, name, extends, doc, fields, examples)
        assert isinstance(subtypes, (tuple, type(None))), type(subtypes)
        self.subtypes = subtypes

    def __repr__(self):
        return 'AstStructDef({!r}, {!r}, {!r})'.format(
            self.name,
            self.extends,
            self.fields,
        )

class AstStructPatch(ASTNode):

    def __init__(self, path, lineno, lexpos, name, fields, examples):
        super().__init__(path, lineno, lexpos)
        self.name = name
        assert isinstance(fields, list)
        self.fields = fields

        assert isinstance(examples, (OrderedDict, type(None))), type(examples)
        self.examples = examples

    def __repr__(self):
        return 'AstStructPatch({!r}, {!r})'.format(
            self.name,
            self.fields,
        )

class AstUnionDef(AstTypeDef):

    def __init__(self, path, lineno, lexpos, name, extends, doc, fields,
                 examples, closed=False):
        """
        Args:
            closed (bool): Set if this is a closed union.

        See AstTypeDef for other constructor args.
        """
        super().__init__(
            path, lineno, lexpos, name, extends, doc, fields, examples)
        self.closed = closed

    def __repr__(self):
        return 'AstUnionDef({!r}, {!r}, {!r}, {!r})'.format(
            self.name,
            self.extends,
            self.fields,
            self.closed,
        )

class AstUnionPatch(ASTNode):

    def __init__(self, path, lineno, lexpos, name, fields, examples, closed):
        super().__init__(path, lineno, lexpos)
        self.name = name
        assert isinstance(fields, list)
        self.fields = fields

        assert isinstance(examples, (OrderedDict, type(None))), type(examples)
        self.examples = examples
        self.closed = closed

    def __repr__(self):
        return 'AstUnionPatch({!r}, {!r}, {!r})'.format(
            self.name,
            self.fields,
            self.closed,
        )

class AstTypeRef(ASTNode):

    def __init__(self, path, lineno, lexpos, name, args, nullable, ns):
        """
        Args:
            name (str): Name of the referenced type.
            args (tuple[list, dict]): Arguments to type.
            nullable (bool): Whether the type is nullable (can be null)
            ns (Optional[str]): Namespace that referred type is a member of.
                If none, then refers to the current namespace.
        """
        super().__init__(path, lineno, lexpos)
        self.name = name
        self.args = args
        self.nullable = nullable
        self.ns = ns

    def __repr__(self):
        return 'AstTypeRef({!r}, {!r}, {!r}, {!r})'.format(
            self.name,
            self.args,
            self.nullable,
            self.ns,
        )

class AstTagRef(ASTNode):

    def __init__(self, path, lineno, lexpos, tag):
        """
        Args:
            tag (str): Name of the referenced type.
        """
        super().__init__(path, lineno, lexpos)
        self.tag = tag

    def __repr__(self):
        return 'AstTagRef({!r})'.format(
            self.tag,
        )

class AstAnnotationRef(ASTNode):

    def __init__(self, path, lineno, lexpos, annotation, ns):
        """
        Args:
            annotation (str): Name of the referenced annotation.
        """
        super().__init__(path, lineno, lexpos)
        self.annotation = annotation
        self.ns = ns

    def __repr__(self):
        return 'AstAnnotationRef({!r}, {!r})'.format(
            self.annotation, self.ns
        )

class AstAnnotationDef(ASTNode):

    def __init__(self, path, lineno, lexpos, name, annotation_type,
                 annotation_type_ns, args, kwargs):
        """
        Args:
            name (str): Name of the defined annotation.
            annotation_type (str): Type of annotation to define.
            annotation_type_ns (Optional[str]): Namespace where the annotation
              type was defined. If None, current namespace or builtin.
            args (str): Arguments to define annotation.
            kwargs (str): Keyword Arguments to define annotation.
        """
        super().__init__(path, lineno, lexpos)
        self.name = name
        self.annotation_type = annotation_type
        self.annotation_type_ns = annotation_type_ns
        self.args = args
        self.kwargs = kwargs

    def __repr__(self):
        return 'AstAnnotationDef({!r}, {!r}, {!r}, {!r}, {!r})'.format(
            self.name,
            self.annotation_type,
            self.annotation_type_ns,
            self.args,
            self.kwargs,
        )

class AstAnnotationTypeDef(ASTNode):

    def __init__(self, path, lineno, lexpos, name, doc, params):
        """
        Args:
            name (str): Name of the defined annotation type.
            doc (str): Docstring for the defined annotation type.
            params (List[AstField]): Parameters that can be passed to the
                annotation type.
        """
        super().__init__(path, lineno, lexpos)
        self.name = name
        self.doc = doc
        self.params = params

    def __repr__(self):
        return 'AstAnnotationTypeDef({!r}, {!r}, {!r})'.format(
            self.name,
            self.doc,
            self.params,
        )

class AstField(ASTNode):
    """
    Represents both a field of a struct and a field of a union.
    TODO(kelkabany): Split this into two different classes.
    """

    def __init__(self, path, lineno, lexpos, name, type_ref):
        """
        Args:
            name (str): The name of the field.
            type_ref (AstTypeRef): The data type of the field.
        """
        super().__init__(path, lineno, lexpos)
        self.name = name
        self.type_ref = type_ref
        self.doc = None
        self.has_default = False
        self.default = None
        self.annotations = []

    def set_doc(self, docstring):
        self.doc = docstring

    def set_default(self, default):
        self.has_default = True
        self.default = default

    def set_annotations(self, annotations):
        self.annotations = annotations

    def __repr__(self):
        return 'AstField({!r}, {!r}, {!r})'.format(
            self.name,
            self.type_ref,
            self.annotations,
        )

class AstVoidField(ASTNode):

    def __init__(self, path, lineno, lexpos, name):
        super().__init__(path, lineno, lexpos)
        self.name = name
        self.doc = None
        self.annotations = []

    def set_doc(self, docstring):
        self.doc = docstring

    def set_annotations(self, annotations):
        self.annotations = annotations

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return 'AstVoidField({!r}, {!r})'.format(
            self.name,
            self.annotations,
        )

class AstSubtypeField(ASTNode):

    def __init__(self, path, lineno, lexpos, name, type_ref):
        super().__init__(path, lineno, lexpos)
        self.name = name
        self.type_ref = type_ref

    def __repr__(self):
        return 'AstSubtypeField({!r}, {!r})'.format(
            self.name,
            self.type_ref,
        )

class AstRouteDef(ASTNode):

    def __init__(self, path, lineno, lexpos, name, version, deprecated,
                 arg_type_ref, result_type_ref, error_type_ref=None):
        super().__init__(path, lineno, lexpos)
        self.name = name
        self.version = version
        self.deprecated = deprecated
        self.arg_type_ref = arg_type_ref
        self.result_type_ref = result_type_ref
        self.error_type_ref = error_type_ref
        self.doc = None
        self.attrs = {}

    def set_doc(self, docstring):
        self.doc = docstring

    def set_attrs(self, attrs):
        self.attrs = attrs

class AstAttrField(ASTNode):

    def __init__(self, path, lineno, lexpos, name, value):
        super().__init__(path, lineno, lexpos)
        self.name = name
        self.value = value

    def __repr__(self):
        return 'AstAttrField({!r}, {!r})'.format(
            self.name,
            self.value,
        )

class AstExample(ASTNode):

    def __init__(self, path, lineno, lexpos, label, text, fields):
        super().__init__(path, lineno, lexpos)
        self.label = label
        self.text = text
        self.fields = fields

    def __repr__(self):
        return 'AstExample({!r}, {!r}, {!r})'.format(
            self.label,
            self.text,
            self.fields,
        )

class AstExampleField(ASTNode):

    def __init__(self, path, lineno, lexpos, name, value):
        super().__init__(path, lineno, lexpos)
        self.name = name
        self.value = value

    def __repr__(self):
        return 'AstExampleField({!r}, {!r})'.format(
            self.name,
            self.value,
        )

class AstExampleRef(ASTNode):

    def __init__(self, path, lineno, lexpos, label):
        super().__init__(path, lineno, lexpos)
        self.label = label

    def __repr__(self):
        return 'AstExampleRef({!r})'.format(self.label)



from collections import OrderedDict
# See <https://github.com/PyCQA/pylint/issues/73>
from packaging.version import Version

from .data_types import (
    doc_unwrap,
    is_alias,
    is_composite_type,
    is_list_type,
    is_nullable_type,
)

_MYPY = False
if _MYPY:
    import typing  # pylint: disable=import-error,useless-suppression

    from .data_types import (  # noqa: F401 # pylint: disable=unused-import
        Alias,
        Annotation,
        AnnotationType,
        DataType,
        List as DataTypeList,
        Nullable,
        Struct,
        UserDefined,
    )

    from stone.frontend.ast import AstRouteDef  # noqa: F401 # pylint: disable=unused-import
    NamespaceDict = typing.Dict[str, 'ApiNamespace']

class Api:
    """
    A full description of an API's namespaces, data types, and routes.
    """
    def __init__(self, version):
        # type: (str) -> None
        self.version = Version(version)
        self.namespaces = OrderedDict()  # type: NamespaceDict
        self.route_schema = None  # type: typing.Optional[Struct]

    def ensure_namespace(self, name):
        # type: (str) -> ApiNamespace
        """
        Only creates a namespace if it hasn't yet been defined.

        :param str name: Name of the namespace.

        :return ApiNamespace:
        """
        if name not in self.namespaces:
            self.namespaces[name] = ApiNamespace(name)
        return self.namespaces[name]

    def normalize(self):
        # type: () -> None
        """
        Alphabetizes namespaces and routes to make spec parsing order mostly
        irrelevant.
        """
        ordered_namespaces = OrderedDict()  # type: NamespaceDict
        # self.namespaces is currently ordered by declaration order.
        for namespace_name in sorted(self.namespaces.keys()):
            ordered_namespaces[namespace_name] = self.namespaces[namespace_name]
        self.namespaces = ordered_namespaces

        for namespace in self.namespaces.values():
            namespace.normalize()

    def add_route_schema(self, route_schema):
        # type: (Struct) -> None
        assert self.route_schema is None
        self.route_schema = route_schema


class _ImportReason:
    """
    Tracks the reason a namespace was imported.
    """

    def __init__(self):
        # type: () -> None
        self.alias = False
        self.data_type = False
        self.annotation = False
        self.annotation_type = False


class ApiNamespace:
    """
    Represents a category of API endpoints and their associated data types.
    """

    def __init__(self, name):
        # type: (typing.Text) -> None
        self.name = name
        self.doc = None                    # type: typing.Optional[typing.Any]
        self.routes = []                   # type: typing.List[ApiRoute]
        # TODO (peichao): route_by_name is deprecated by routes_by_name and should be removed.
        self.route_by_name = {}            # type: typing.Dict[typing.Text, ApiRoute]
        self.routes_by_name = {}           # type: typing.Dict[typing.Text, ApiRoutesByVersion]
        self.data_types = []               # type: typing.List[UserDefined]
        self.data_type_by_name = {}        # type: typing.Dict[str, UserDefined]
        self.aliases = []                  # type: typing.List[Alias]
        self.alias_by_name = {}            # type: typing.Dict[str, Alias]
        self.annotations = []              # type: typing.List[Annotation]
        self.annotation_by_name = {}       # type: typing.Dict[str, Annotation]
        self.annotation_types = []         # type: typing.List[AnnotationType]
        self.annotation_type_by_name = {}  # type: typing.Dict[str, AnnotationType]
        self._imported_namespaces = {}     # type: typing.Dict[ApiNamespace, _ImportReason]

    def add_doc(self, docstring):
        # type: (typing.Any) -> None
        """Adds a docstring for this namespace.

        The input docstring is normalized to have no leading whitespace and
        no trailing whitespace except for a newline at the end.

        If a docstring already exists, the new normalized docstring is appended
        to the end of the existing one with two newlines separating them.
        """
        assert isinstance(docstring, str), type(docstring)
        normalized_docstring = doc_unwrap(docstring) + '\n'
        if self.doc is None:
            self.doc = normalized_docstring
        else:
            self.doc += normalized_docstring

    def add_route(self, route):
        # type: (ApiRoute) -> None
        self.routes.append(route)
        if route.version == 1:
            self.route_by_name[route.name] = route
        if route.name not in self.routes_by_name:
            self.routes_by_name[route.name] = ApiRoutesByVersion()
        self.routes_by_name[route.name].at_version[route.version] = route

    def add_data_type(self, data_type):
        # type: (UserDefined) -> None
        self.data_types.append(data_type)
        self.data_type_by_name[data_type.name] = data_type

    def add_alias(self, alias):
        # type: (Alias) -> None
        self.aliases.append(alias)
        self.alias_by_name[alias.name] = alias

    def add_annotation(self, annotation):
        # type: (Annotation) -> None
        self.annotations.append(annotation)
        self.annotation_by_name[annotation.name] = annotation

    def add_annotation_type(self, annotation_type):
        # type: (AnnotationType) -> None
        self.annotation_types.append(annotation_type)
        self.annotation_type_by_name[annotation_type.name] = annotation_type

    def add_imported_namespace(self,
                               namespace,
                               imported_alias=False,
                               imported_data_type=False,
                               imported_annotation=False,
                               imported_annotation_type=False):
        # type: (ApiNamespace, bool, bool, bool, bool) -> None
        """
        Keeps track of namespaces that this namespace imports.

        Args:
            namespace (Namespace): The imported namespace.
            imported_alias (bool): Set if this namespace references an alias
                in the imported namespace.
            imported_data_type (bool): Set if this namespace references a
                data type in the imported namespace.
            imported_annotation (bool): Set if this namespace references a
                annotation in the imported namespace.
            imported_annotation_type (bool): Set if this namespace references an
                annotation in the imported namespace, possibly indirectly (by
                referencing an annotation elsewhere that has this type).
        """
        assert self.name != namespace.name, \
            'Namespace cannot import itself.'
        reason = self._imported_namespaces.setdefault(namespace, _ImportReason())
        if imported_alias:
            reason.alias = True
        if imported_data_type:
            reason.data_type = True
        if imported_annotation:
            reason.annotation = True
        if imported_annotation_type:
            reason.annotation_type = True

    def linearize_data_types(self):
        # type: () -> typing.List[UserDefined]
        """
        Returns a list of all data types used in the namespace. Because the
        inheritance of data types can be modeled as a DAG, the list will be a
        linearization of the DAG. It's ideal to generate data types in this
        order so that composite types that reference other composite types are
        defined in the correct order.
        """
        linearized_data_types = []
        seen_data_types = set()  # type: typing.Set[UserDefined]

        def add_data_type(data_type):
            # type: (UserDefined) -> None
            if data_type in seen_data_types:
                return
            elif data_type.namespace != self:
                # We're only concerned with types defined in this namespace.
                return
            if is_composite_type(data_type) and data_type.parent_type:
                add_data_type(data_type.parent_type)
            linearized_data_types.append(data_type)
            seen_data_types.add(data_type)

        for data_type in self.data_types:
            add_data_type(data_type)

        return linearized_data_types

    def linearize_aliases(self):
        # type: () -> typing.List[Alias]
        """
        Returns a list of all aliases used in the namespace. The aliases are
        ordered to ensure that if they reference other aliases those aliases
        come earlier in the list.
        """
        linearized_aliases = []
        seen_aliases = set()  # type: typing.Set[Alias]

        def add_alias(alias):
            # type: (Alias) -> None
            if alias in seen_aliases:
                return
            elif alias.namespace != self:
                return
            if is_alias(alias.data_type):
                add_alias(alias.data_type)
            linearized_aliases.append(alias)
            seen_aliases.add(alias)

        for alias in self.aliases:
            add_alias(alias)

        return linearized_aliases

    def get_route_io_data_types(self):
        # type: () -> typing.List[UserDefined]
        """
        Returns a list of all user-defined data types that are referenced as
        either an argument, result, or error of a route. If a List or Nullable
        data type is referenced, then the contained data type is returned
        assuming it's a user-defined type.
        """
        data_types = set()  # type: typing.Set[UserDefined]
        for route in self.routes:
            data_types |= self.get_route_io_data_types_for_route(route)
        return sorted(data_types, key=lambda dt: dt.name)

    def get_route_io_data_types_for_route(self, route):
        # type: (ApiRoute) -> typing.Set[UserDefined]
        """
        Given a route, returns a set of its argument/result/error datatypes.
        """
        data_types = set()  # type: typing.Set[UserDefined]
        for dtype in (route.arg_data_type, route.result_data_type, route.error_data_type):
            while is_list_type(dtype) or is_nullable_type(dtype):
                data_list_type = dtype  # type: typing.Any
                dtype = data_list_type.data_type
            if is_composite_type(dtype) or is_alias(dtype):
                data_user_type = dtype  # type: typing.Any
                data_types.add(data_user_type)
        return data_types

    def get_imported_namespaces(self,
                                must_have_imported_data_type=False,
                                consider_annotations=False,
                                consider_annotation_types=False):
        # type: (bool, bool, bool) -> typing.List[ApiNamespace]
        """
        Returns a list of Namespace objects. A namespace is a member of this
        list if it is imported by the current namespace and a data type is
        referenced from it. Namespaces are in ASCII order by name.

        Args:
            must_have_imported_data_type (bool): If true, result does not
                include namespaces that were not imported for data types.
            consider_annotations (bool): If false, result does not include
                namespaces that were only imported for annotations
            consider_annotation_types (bool): If false, result does not
                include namespaces that were only imported for annotation types.

        Returns:
            List[Namespace]: A list of imported namespaces.
        """
        imported_namespaces = []
        for imported_namespace, reason in self._imported_namespaces.items():
            if must_have_imported_data_type and not reason.data_type:
                continue
            if (not consider_annotations) and not (
                    reason.data_type or reason.alias or reason.annotation_type
            ):
                continue
            if (not consider_annotation_types) and not (
                    reason.data_type or reason.alias or reason.annotation
            ):
                continue

            imported_namespaces.append(imported_namespace)
        imported_namespaces.sort(key=lambda n: n.name)
        return imported_namespaces

    def get_namespaces_imported_by_route_io(self):
        # type: () -> typing.List[ApiNamespace]
        """
        Returns a list of Namespace objects. A namespace is a member of this
        list if it is imported by the current namespace and has a data type
        from it referenced as an argument, result, or error of a route.
        Namespaces are in ASCII order by name.
        """
        namespace_data_types = sorted(self.get_route_io_data_types(),
                                      key=lambda dt: dt.name)
        referenced_namespaces = set()
        for data_type in namespace_data_types:
            if data_type.namespace != self:
                referenced_namespaces.add(data_type.namespace)
        return sorted(referenced_namespaces, key=lambda n: n.name)

    def normalize(self):
        # type: () -> None
        """
        Sorts routes to make route declaration order irrelevant.
        """
        self.routes.sort()
        self.data_types.sort(key=lambda data_type: data_type.name)
        self.aliases.sort(key=lambda alias: alias.name)
        self.annotations.sort(key=lambda annotation: annotation.name)

    def __repr__(self):
        # type: () -> str
        return 'ApiNamespace({!r})'.format(self.name)


class ApiRoute:
    """
    Represents an API endpoint.
    """

    def __init__(self,
                 name,
                 version,
                 ast_node):
        # type: (typing.Text, int, typing.Optional[AstRouteDef]) -> None
        """
        :param str name: Designated name of the endpoint.
        :param int version: Designated version of the endpoint.
        :param ast_node: Raw route definition from the parser.
        """
        self.name = name
        self.version = version
        self._ast_node = ast_node

        # These attributes are set later by set_attributes()
        self.deprecated = None  # type: typing.Optional[DeprecationInfo]
        self.raw_doc = None  # type: typing.Optional[typing.Text]
        self.doc = None  # type: typing.Optional[typing.Text]
        self.arg_data_type = None  # type: typing.Optional[DataType]
        self.result_data_type = None  # type: typing.Optional[DataType]
        self.error_data_type = None  # type: typing.Optional[DataType]
        self.attrs = None  # type: typing.Optional[typing.Mapping[typing.Text, typing.Any]]

    def set_attributes(self, deprecated, doc, arg_data_type, result_data_type,
                       error_data_type, attrs):
        """
        Converts a forward reference definition of a route into a full
        definition.

        :param DeprecationInfo deprecated: Set if this route is deprecated.
        :param str doc: Description of the endpoint.
        :type arg_data_type: :class:`stone.data_type.DataType`
        :type result_data_type: :class:`stone.data_type.DataType`
        :type error_data_type: :class:`stone.data_type.DataType`
        :param dict attrs: Map of string keys to values that are either int,
            float, bool, str, or None. These are the route attributes assigned
            in the spec.
        """
        self.deprecated = deprecated
        self.raw_doc = doc
        self.doc = doc_unwrap(doc)
        self.arg_data_type = arg_data_type
        self.result_data_type = result_data_type
        self.error_data_type = error_data_type
        self.attrs = attrs

    def name_with_version(self):
        """
        Get user-friendly representation of the route.

        :return: Route name with version suffix. The version suffix is omitted for version 1.
        """
        if self.version == 1:
            return self.name
        else:
            return '{}:{}'.format(self.name, self.version)

    def __repr__(self):
        return 'ApiRoute({})'.format(self.name_with_version())

    def _compare(self, lhs, rhs):
        if not isinstance(lhs, ApiRoute):
            raise TypeError("Expected ApiRoute for object: {}".format(lhs))
        if not isinstance(rhs, ApiRoute):
            raise TypeError("Expected ApiRoute for object: {}".format(rhs))

        if (lhs.name, lhs.version) < (rhs.name, rhs.version):
            return -1
        elif (lhs.name, lhs.version) > (rhs.name, rhs.version):
            return 1
        else:
            return 0

    def __lt__(self, other):
        return self._compare(self, other) < 0
    def __gt__(self, other):
        return self._compare(self, other) > 0
    def __eq__(self, other):
        return self._compare(self, other) == 0
    def __le__(self, other):
        return self._compare(self, other) <= 0
    def __ge__(self, other):
        return self._compare(self, other) >= 0
    def __ne__(self, other):
        return self._compare(self, other) != 0
    def __hash__(self):
        return hash((
            self.name,
            self.version,
            self._ast_node,
            self.deprecated,
            self.raw_doc,
            self.doc,
            self.arg_data_type,
            self.result_data_type,
            self.error_data_type,
            id(self.attrs)
        ))


class DeprecationInfo:

    def __init__(self, by=None):
        # type: (typing.Optional[ApiRoute]) -> None
        """
        :param ApiRoute by: The route that replaces this deprecated one.
        """
        assert by is None or isinstance(by, ApiRoute), repr(by)
        self.by = by

class ApiRoutesByVersion:
    """
    Represents routes of different versions for a common name.
    """

    def __init__(self):
        # type: () -> None
        """
        :param at_version: The dict mapping a version number to a route.
        """
        self.at_version = {}  # type: typing.Dict[int, ApiRoute]



import re

_split_words_capitalization_re = re.compile(
    '^[a-z0-9]+|[A-Z][a-z0-9]+|[A-Z]+(?=[A-Z][a-z0-9])|[A-Z]+$'
)
_split_words_dashes_re = re.compile('[-_/]+')


def split_words(name):
    """
    Splits name based on capitalization, dashes, and underscores.
        Example: 'GetFile' -> ['Get', 'File']
        Example: 'get_file' -> ['get', 'file']
    """
    all_words = []
    for word in re.split(_split_words_dashes_re, name):
        vals = _split_words_capitalization_re.findall(word)
        if vals:
            all_words.extend(vals)
        else:
            all_words.append(word)
    return all_words


def fmt_camel(name):
    """
    Converts name to lower camel case. Words are identified by capitalization,
    dashes, and underscores.
    """
    words = split_words(name)
    assert len(words) > 0
    first = words.pop(0).lower()
    return first + ''.join([word.capitalize() for word in words])


def fmt_dashes(name):
    """
    Converts name to words separated by dashes. Words are identified by
    capitalization, dashes, and underscores.
    """
    return '-'.join([word.lower() for word in split_words(name)])


def fmt_pascal(name):
    """
    Converts name to pascal case. Words are identified by capitalization,
    dashes, and underscores.
    """
    return ''.join([word.capitalize() for word in split_words(name)])


def fmt_underscores(name):
    """
    Converts name to words separated by underscores. Words are identified by
    capitalization, dashes, and underscores.
    """
    return '_'.join([word.lower() for word in split_words(name)])
